@@ -24,6 +24,9 @@
 
  ********************************************************/
 
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 #include <stdio.h>
 #include <stdlib.h>
 #include <locale.h>
@@ -35,6 +38,7 @@
 #include <X11/extensions/XKBfile.h>
 #include <X11/extensions/XKBconfig.h>
 #include <X11/extensions/XKBrules.h>
+#include <X11/extensions/Xrandr.h>
 
 #ifndef PATH_MAX
 #ifdef MAXPATHLEN
@@ -129,7 +133,6 @@ static settings_t settings = {
 
 static XkbConfigRtrnRec cfgResult;
 
-static XkbRF_RulesPtr rules = NULL;
 static XkbRF_VarDefsRec rdefs;
 
 static Bool clearOptions = False;
@@ -170,26 +173,26 @@ static int deviceSpec = XkbUseCoreKbd;
 #define ERR2(s,a,b)     fprintf(stderr,s,a,b)
 #define ERR3(s,a,b,c)   fprintf(stderr,s,a,b,c)
 
-#define OOM(ptr)        do { if ((ptr) == NULL) { ERR("Out of memory.\n"); exit(-1); } } while (0)
+#define OOM(ptr)        do { if ((ptr) == NULL) { ERR("Out of memory.\n"); exit(1); } } while (0)
 
 /***====================================================================***/
 
-Bool addToList(list_t *list, const char *newVal);
-void usage(int argc, char **argv);
-void dumpNames(Bool wantRules, Bool wantCNames);
-void trySetString(setting_t * setting, char *newVal, enum source src);
-Bool setOptString(int *arg, int argc, char **argv, setting_t *setting, enum source src);
-int parseArgs(int argc, char **argv);
-Bool getDisplay(int argc, char **argv);
-Bool getServerValues(void);
-FILE *findFileInPath(char *name);
-Bool addStringToOptions(char *opt_str, list_t *opts);
-char *stringFromOptions(char *orig, list_t *newOpts);
-Bool applyConfig(char *name);
-XkbRF_RulesPtr tryLoadRules(char *name, char *locale, Bool wantDesc, Bool wantRules);
-Bool applyRules(void);
-Bool applyComponentNames(void);
-void printKeymap(void);
+static Bool addToList(list_t *list, const char *newVal);
+static void usage(int argc, char **argv);
+static void dumpNames(Bool wantRules, Bool wantCNames);
+static void trySetString(setting_t * setting, char *newVal, enum source src);
+static Bool setOptString(int *arg, int argc, char **argv, setting_t *setting, enum source src);
+static int parseArgs(int argc, char **argv);
+static Bool getDisplay(int argc, char **argv);
+static Bool getServerValues(void);
+static FILE *findFileInPath(char *name);
+static Bool addStringToOptions(char *opt_str, list_t *opts);
+static char *stringFromOptions(char *orig, list_t *newOpts);
+static Bool applyConfig(char *name);
+static XkbRF_RulesPtr tryLoadRules(char *name, char *locale, Bool wantDesc, Bool wantRules);
+static Bool applyRules(void);
+static Bool applyComponentNames(void);
+static void printKeymap(void);
 
 /***====================================================================***/
 
@@ -198,7 +201,7 @@ void printKeymap(void);
     Otherwise newVal is added to the end of the list (if it is not present in the list yet).
 */
 
-Bool
+static Bool
 addToList(list_t *list, const char *newVal)
 {
     register int i;
@@ -234,7 +237,7 @@ addToList(list_t *list, const char *newVal)
 
 /***====================================================================***/
 
-void
+static void
 usage(int argc, char **argv)
 {
     MSG1(
@@ -246,7 +249,7 @@ usage(int argc, char **argv)
         "  -device <deviceid>  Specifies the device ID to use\n"
         "  -display <dpy>      Specifies display to use\n"
         "  -geometry <name>    Specifies geometry component name\n"
-        "  -I<dir>             Add <dir> to list of directories to be used\n"
+        "  -I <dir>            Add <dir> to list of directories to be used\n"
         "  -keycodes <name>    Specifies keycodes component name\n"
         "  -keymap <name>      Specifies name of keymap to load\n"
         "  -layout <name>      Specifies layout used to choose component names\n"
@@ -258,14 +261,14 @@ usage(int argc, char **argv)
         "  -symbols <name>     Specifies symbols component name\n"
         "  -synch              Synchronize request with X server\n"
         "  -types <name>       Specifies types component name\n"
-        "  -v[erbose] [<lvl>]  Sets verbosity (1..10).  Higher values yield\n"
-        "                      more messages\n"
+        "  -v[erbose] [<lvl>]  Sets verbosity (1..10); higher values yield more messages\n"
+        "  -version            Print the program's version number\n"
         "  -variant <name>     Specifies layout variant used to choose component names\n",
         argv[0]
     );
 }
 
-void
+static void
 dumpNames(Bool wantRules, Bool wantCNames)
 {
     if (wantRules)
@@ -312,7 +315,7 @@ dumpNames(Bool wantRules, Bool wantCNames)
  *
  * @param which What value is it (one of RULES_NDX, CONFIG_NDX, ...)
  */
-void
+static void
 trySetString(setting_t *setting, char *newVal, enum source src)
 {
     if (setting->value != NULL)
@@ -338,7 +341,7 @@ trySetString(setting_t *setting, char *newVal, enum source src)
     return;
 }
 
-Bool
+static Bool
 setOptString(int *arg, int argc, char **argv, setting_t *setting, enum source src)
 {
     int ndx;
@@ -383,7 +386,7 @@ setOptString(int *arg, int argc, char **argv, setting_t *setting, enum source sr
  * Return True on success or False if an unrecognized option has been
  * specified.
  */
-int
+static int
 parseArgs(int argc, char **argv)
 {
     int i;
@@ -416,7 +419,7 @@ parseArgs(int argc, char **argv)
                 deviceSpec = atoi(argv[i]); /* only allow device IDs, not names */
             } else {
                 usage(argc, argv);
-                exit(-1);
+                exit(1);
             }
         }
         else if (streq(argv[i], "-display"))
@@ -473,6 +476,11 @@ parseArgs(int argc, char **argv)
             synch = True;
         else if (streq(argv[i], "-types"))
             ok = setOptString(&i, argc, argv, &settings.types, FROM_CMD_LINE);
+        else if (streq(argv[i], "-version"))
+        {
+            MSG1("setxkbmap %s\n", PACKAGE_VERSION);
+            exit(0);
+        }
         else if (streq(argv[i], "-verbose") || (streq(argv[i], "-v")))
         {
             if ((i < argc - 1) && (isdigit(argv[i + 1][0])))
@@ -532,7 +540,7 @@ parseArgs(int argc, char **argv)
  *
  * @return True on success or False otherwise.
  */
-Bool
+static Bool
 getDisplay(int argc, char **argv)
 {
     int major, minor, why;
@@ -547,7 +555,7 @@ getDisplay(int argc, char **argv)
         if (settings.display.value == NULL)
             settings.display.value = getenv("DISPLAY");
         if (settings.display.value == NULL)
-            settings.display.value = "default display";
+            settings.display.value = (char *) "default display";
         switch (why)
         {
         case XkbOD_BadLibraryVersion:
@@ -588,7 +596,7 @@ getDisplay(int argc, char **argv)
  *
  * @return True.
  */
-Bool
+static Bool
 getServerValues(void)
 {
     XkbRF_VarDefsRec vd;
@@ -597,9 +605,9 @@ getServerValues(void)
     if (!XkbRF_GetNamesProp(dpy, &tmp, &vd) || !tmp)
     {
         VMSG1(3, "Couldn't interpret %s property\n", _XKB_RF_NAMES_PROP_ATOM);
-        tmp = DFLT_XKB_RULES_FILE;
-        vd.model = DFLT_XKB_MODEL;
-        vd.layout = DFLT_XKB_LAYOUT;
+        tmp = (char *) DFLT_XKB_RULES_FILE;
+        vd.model = (char *) DFLT_XKB_MODEL;
+        vd.layout = (char *) DFLT_XKB_LAYOUT;
         vd.variant = NULL;
         vd.options = NULL;
         VMSG3(3, "Use defaults: rules - '%s' model - '%s' layout - '%s'\n",
@@ -623,7 +631,7 @@ getServerValues(void)
 
 /***====================================================================***/
 
-FILE *
+static FILE *
 findFileInPath(char *name)
 {
     register int i;
@@ -657,7 +665,7 @@ findFileInPath(char *name)
 
 /***====================================================================***/
 
-Bool
+static Bool
 addStringToOptions(char *opt_str, list_t *opts)
 {
     char *tmp, *str, *next;
@@ -681,10 +689,11 @@ addStringToOptions(char *opt_str, list_t *opts)
 
 /***====================================================================***/
 
-char *
+static char *
 stringFromOptions(char *orig, list_t *newOpts)
 {
-    int len, i, nOut;
+    size_t len;
+    int i, nOut;
 
     if (orig)
         len = strlen(orig) + 1;
@@ -727,19 +736,22 @@ stringFromOptions(char *orig, list_t *newOpts)
 
 /***====================================================================***/
 
-Bool
+static Bool
 applyConfig(char *name)
 {
     FILE *fp;
     Bool ok;
 
     if ((fp = findFileInPath(name)) == NULL)
+    {
+        ERR1("Couldn't find configuration file \"%s\"\n", name);
         return False;
+    }
     ok = XkbCFParse(fp, XkbCFDflts, NULL, &cfgResult);
     fclose(fp);
     if (!ok)
     {
-        ERR1("Couldn't find configuration file \"%s\"\n", name);
+        ERR1("Couldn't parse configuration file \"%s\"\n", name);
         return False;
     }
     if (cfgResult.rules_file)
@@ -805,7 +817,7 @@ applyConfig(char *name)
     return True;
 }
 
-XkbRF_RulesPtr
+static XkbRF_RulesPtr
 tryLoadRules(char *name, char *locale, Bool wantDesc, Bool wantRules)
 {
     XkbRF_RulesPtr rules = NULL;
@@ -824,11 +836,12 @@ tryLoadRules(char *name, char *locale, Bool wantDesc, Bool wantRules)
  *
  * @return True on success or false otherwise.
  */
-Bool
+static Bool
 applyRules(void)
 {
     int i;
     char *rfName;
+    XkbRF_RulesPtr rules = NULL;
 
     if (settings.model.src || settings.layout.src || settings.variant.src
         || options.item)
@@ -849,7 +862,7 @@ applyRules(void)
         if (settings.rules.src)
             rfName = settings.rules.value;
         else
-            rfName = DFLT_XKB_RULES_FILE;
+            rfName = (char *) DFLT_XKB_RULES_FILE;
 
         if (rfName[0] == '/')
         {
@@ -976,7 +989,7 @@ checkName(char *name, const char *string)
     return ret;
 }
 
-void
+static void
 printKeymap(void)
 {
     MSG("xkb_keymap {\n");
@@ -993,7 +1006,7 @@ printKeymap(void)
     MSG("};\n");
 }
 
-Bool
+static Bool
 applyComponentNames(void)
 {
     if (!checkName(settings.types.value, "types"))
@@ -1017,13 +1030,15 @@ applyComponentNames(void)
     /* Upload the new description to the server. */
     if (dpy && !print && !query)
     {
-        XkbComponentNamesRec cmdNames;
-        cmdNames.types = settings.types.value;
-        cmdNames.compat = settings.compat.value;
-        cmdNames.symbols = settings.symbols.value;
-        cmdNames.keycodes = settings.keycodes.value;
-        cmdNames.geometry = settings.geometry.value;
-        cmdNames.keymap = settings.keymap.value;
+        XkbComponentNamesRec cmdNames = {
+            .keymap = settings.keymap.value,
+            .keycodes = settings.keycodes.value,
+            .types = settings.types.value,
+            .compat = settings.compat.value,
+            .symbols = settings.symbols.value,
+            .geometry = settings.geometry.value
+        };
+
         xkb = XkbGetKeyboardByName(dpy, deviceSpec, &cmdNames,
                                    XkbGBN_AllComponentsMask,
                                    XkbGBN_AllComponentsMask &
@@ -1053,23 +1068,81 @@ applyComponentNames(void)
     return True;
 }
 
+static Bool
+is_xwayland(void)
+{
+    /* Code copied from xisxwayland.c */
+    Bool rc = False;
+    XRRScreenResources *resources = NULL;
+    XRROutputInfo *output = NULL;
+    int opcode, event_base, error_base, major, minor;
+
+    /* With Xwayland 23.1 and above, the definitive way of checking for an
+     * Xwayland server is to check for the "XWAYLAND" extension.
+     */
+    if (XQueryExtension(dpy, "XWAYLAND", &opcode, &event_base, &error_base)) {
+        rc = True;
+        goto out;
+    }
+
+    /* For previous versions of Xwayland, there is no definitive way of checking
+     * for an Xwayland server, but the two working methods are:
+     * - RandR output names in Xwayland are XWAYLAND0, XWAYLAND1, etc.
+     * - XI devices are xwayland-pointer:10, xwayland-keyboard:11
+     * Let's go with the XRandR check here because it's slightly less
+     * code to write.
+     */
+
+    if (!XRRQueryExtension(dpy, &event_base, &error_base) ||
+            !XRRQueryVersion(dpy, &major, &minor)) {
+        /* e.g. Xnest, but definitely not Xwayland */
+        goto out;
+    }
+    if (major < 1 || (major == 1 && minor < 3)) {
+        goto out;
+    }
+
+    resources = XRRGetScreenResourcesCurrent(dpy, DefaultRootWindow(dpy));
+    if (!resources || resources->noutput < 1) {
+        goto out;
+    }
+
+    output = XRRGetOutputInfo(dpy, resources, resources->outputs[0]);
+    if (!output) {
+        goto out;
+    }
+
+    if (strncmp(output->name, "XWAYLAND", 8) == 0)
+        rc = True;
+
+    XRRFreeOutputInfo(output);
+out:
+    if (resources)
+        XRRFreeScreenResources(resources);
+
+    return rc;
+}
 
 int
 main(int argc, char **argv)
 {
     if ((!parseArgs(argc, argv)) || (!getDisplay(argc, argv)))
-        exit(-1);
+        exit(1);
+
+    if (is_xwayland())
+	    ERR("WARNING: Running setxkbmap against an Xwayland server\n");
+
     settings.locale.value = setlocale(LC_ALL, settings.locale.value);
     settings.locale.src = FROM_SERVER;
     VMSG1(7, "locale is %s\n", settings.locale.value);
     if (dpy)
         getServerValues();
     if (settings.config.value && (!applyConfig(settings.config.value)))
-        exit(-3);
+        exit(3);
     if (!applyRules())
-        exit(-4);
+        exit(4);
     if (!applyComponentNames())
-        exit(-5);
+        exit(5);
     if (dpy)
         XCloseDisplay(dpy);
     exit(0);
